<?php

/***
 * 
 * k.Bot - modular irc bot
 * Copyright (C) 2004 Kore Nordmann - http://www.kore-nordmann.de (German)
 * 
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or (at 
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA. 
 * 
 * You may contact the author by mail:
 * kbot@kore-nordmann.de
 *
 * Or write to:
 * Kore Nordmann
 * Stockumer Strasse 404
 * 44227 Dortmund
 * Germany
 *
 * The latest version of k.Bot can be obtained from:
 * http://www.kore-nordmann.de/
 *
 **/

// Konstanten - bitte NICHT aendern
define('IRC_GETMODE',		1);
define('IRC_LOSEMODE',		2);
define('IRC_JOIN',			3);
define('IRC_PART',			4);
define('IRC_KICK',			5);
define('IRC_QUIT',			6);
define('IRC_NICK',			7);
define('IRC_COMMAND',		8);
define('IRC_MSG',			9);
define('IRC_CONNECT',		10);

define('IRC_NOMODE',		0);
define('IRC_VOICE',			1);
define('IRC_HALFOP',		2);
define('IRC_OP',			4);
define('IRC_ADMIN',			8);
define('IRC_FOUNDER',		16);

define('KBOT_VERSION',		'k.Bot 3.0');

class irc {
	
	private	$nick;			// Nick
	private	$ident;			// Ident
	private $name;			// Name
	
	private	$server;		// Serverliste
	private	$port;			// Portliste
	private $actualServer;	// Aktueller Server
	
	private $messages;		// Messagequeue
	private $lastMessage;	// Letzte Nachricht vom Server
	private $actualNick;	// Aktueller Nick
	private $log;			// Log
	private	$connection;	// Connectionhandle
	private	$logSize;		// Laenge des Logs
	
	/*
	 * Signalarray
	 *
	 * Die geparsten Nachrichten werden in den Array gepusht, um dann ausgelesen
	 * werden zu koennen.
	 *
	 * Jedes Signal ist ein Array, an Index 0 steht der Typ (siehe Konstanten).
	 * Die weiteren Indizes sind vom Typ anhaengig.
	 */
	public	$signals;
	
	/**
	* @return irc
	* @param string $nick	Nick
	* @param string $ident	Ident
	* @param string $name	Name
	* @desc	Konstruktor der IRC-Klasse
	*/
	public function __construct(		$nick = 'kBot',
										$ident = 'kbot',
										$name = 'k.Bot 3.0') {
		$this->nick = (string) $nick;
		$this->actualNick = $this->nick;
		$this->ident = (string) $ident;
		$this->name = (string) $name;
		
		$this->server = array();
		$this->port = array();
		$this->actualServer = 0;
		
		$this->messages = array();
		$this->lastMessage = time();
		$this->signals = array();
		
		$this->log = array();
		$this->logSize = KBOT_LOG_SIZE;
	}
	
	/**
	* @return void
	* @desc Destruktor der IRC-Klasse
	*/
	public function __destruct() {
		$this->quit('Bot destroyed.');
		@fclose($this->connection);
		echo "<pre>\n" . $this->showLog() . "</pre>\n";
	}
	
	/**
	* @return void
	* @param string $string	Logbucheintrag
	* @desc	Fuegt einen Eintrag in das Logbuch ein
	*/
	public function log(				$string) {
		echo time() . ': ' . $string . "\n";
		$this->log[] = time() . ': ' . $string;
		$this->log = array_slice($this->log, -$this->logSize);
	}
	
	/**
	* @return void
	* @param string $string	Nick
	* @desc Setzt den angegeben Nick als Nickname des Bots
	*/
	public function setNick(			$string) {
		if ($string !== $this->nick) $this->nick = (string) $string;
		if ($this->nick !== $this->actualNick) $this->nick($this->nick);
	}
	
	/**
	* @return void
	* @param string	$string	IRC-Kommando
	* @desc Setzt das IRC-Kommando auf dem Server ab
	*/
	private function message(			$string) {
		if (!$this->connection) {
			$this->connect();
		} else {
			if (!@fputs($this->connection, trim($string) . "\r\n")) $this->disconnect();
		}
		usleep(10000);
	}
	
	/**
	* @return boolean
	* @param string $server	Host
	* @param int	$port	Port
	* @desc Fuegt einen Server in die Serverliste ein
	*/
	public function addServer(			$server,
										$port = 6667) {
		if (!$ip = @gethostbyname($server)) {
			$this->log('Kann Namen des Servers "' . $server . '" nicht aufloesen.');
			return false; 
		} else {
			$this->server[] = $ip;
		}
		$this->port[] = (int) $port;
		return true;
	}
	
	/**
	* @return boolean
	* @desc	Versucht die Verbindung zum naechsten Server herzustellen
	*/
	public function connect() {
		// Ob Server in der Liste sind
		if (!isset($this->server[$this->actualServer])) {
			$this->log('Kann keine Verbindung aufbauen, bevor Server eingefuegt wurden.');
			return false;
		}
		
		// Logge den verbindungsversuch
		$this->log('Verbinde mich zu irc://' . $this->server[$this->actualServer] . ':' . $this->port[$this->actualServer] . '/.');
		// Verbindungsaufbau
		$this->connection = @fsockopen($this->server[$this->actualServer], $this->port[$this->actualServer]);
		usleep(500000);
		
		// Wenn keine Verbindung, dann den naechsten Server versuchen
		if (!$this->connection) {
			$this->log('Verbindungsaufbau gescheitert, versuche naechsten Server in der Liste.');
			$this->actualServer = ($this->actualServer + 1) % count($this->server);
			sleep(KBOT_RECONNECT);
			$this->connect();
			return false;
		}
		
		$this->lastMessage = time();
		
		// Eigenschaften des Sockets setzen
		socket_set_timeout($this->connection, 0.5);
		socket_set_blocking($this->connection, false);
		
		// Anmelden auf dem Server
		$this->login();
		
		return true;
	}
	
	/**
	* @return void
	* @desc Meldet sich beim Server an
	*/
	private function login() {
		$this->log('Login');
		$this->message('USER ' . $this->ident . ' foo foo :' . $this->name);
		usleep(500000);
		$this->readMessages();
		$this->nick($this->nick);
	}
	
	/**
	* @return void
	* @param string $nick	Nick
	* @desc Nick-Kommando
	*/
	public function nick(				$nick) {
		$this->message('NICK ' . $nick);
		$this->readMessages();
	}
	
	/**
	* @return void
	* @param string	$string	Quit-Nachricht
	* @desc Quit-kommando
	*/
	public function quit(				$string) {
		$this->message('QUIT :' . $string);
	}
	
	/**
	* @return void
	* @param string	$channel Channel
	* @desc Join-Kommando
	*/
	public function join(				$channel) {
		$this->message('JOIN ' . $channel);
		$this->readMessages();
	}

	/**
	* @return void
	* @param string $channel Channel
	* @desc Part-Kommando
	*/
	public function part(				$channel) {
		$this->message('PART ' . $channel);
		$this->readMessages();
	}
	
	/**
	* @return void
	* @param string $channel	Channel
	* @param string $user		User
	* @param string $comment	Kommentar
	* @desc Kick-Kommando
	*/
	public function kick(				$channel,
										$user,
										$comment) {
		$this->message('KICK ' . $channel . ' ' . $user . ' :' . $comment);
		$this->log('KICK ' . $channel . ' ' . $user . ' :' . $comment);
		$this->readMessages();
	}

	/**
	* @return void
	* @param string $channel	Channel
	* @param string $modes		Modes
	* @param string $user		User
	* @desc Mode-Kommando
	*/
	public function mode(				$channel,
										$modes,
										$user = '') {
		$this->message('MODE ' . $channel . ' ' . $modes . ' ' . $user);
		$this->readMessages();
	}

	/**
	* @return void
	* @param string $channel	Channel
	* @param string $string		Nachricht
	* @desc Nachricht schreiben
	*/
	public function msg(				$channel,
										$string) {
		// Zeilenumbrueche und zu lange Nachrichten behandeln
		$string = wordwrap($string, 400);
		$strings = preg_split("(\n|\r\n|\r)", $string);
		foreach ($strings as $string) $this->message('PRIVMSG ' . $channel . ' :' . $string);
		$this->readMessages();
	}
	
	/**
	* @return void
	* @param string $channel	Channel
	* @param string $string		Nachricht
	* @desc Notice schreiben
	*/
	public function notice(				$channel,
										$string) {
		$this->message('NOTICE ' . $channel . ' :' . $string);
		$this->readMessages();
	}
	
	/**
	* @return array
	* @desc Die letzten Logbucheintraege zurueckgeben
	*/
	public function showLog() {
		$log = '';
		foreach ($this->log as $time => $line) {
			$log .= $line . "\n";
		}
		return $log;
	}
	
	/**
	* @return boolean
	* @desc Test, ob die Verbindung abgelaufen ist
	*/
	private function timedOut() {
		if (($this->lastMessage + KBOT_TIMEOUT) < time()) {
			$this->log('Verbindung abgelaufen - letzte Nachricht: ' . date('H:i.s', $this->lastMessage) . '.');
			return true;
		} else {
			return false;
		}
	}
	
	/**
	* @return void
	* @desc Liest die naechsten anstehenden Nachrichten vom Socket
	*/
	private function readMessages() {
		while ($buffer = @fgets($this->connection)) {
			$this->lastMessage = time();
			
			if (substr($buffer, 0, 4) == 'PING') {
				// Sofoertiger PONG, falls PING
				$this->message('PONG ' . trim(substr($buffer, 5)));
			} else {
				// Alles andere parsen und in die messagequeue einfuegen
				if (!empty($buffer)) $this->messages[] = $this->parseMessage(trim($buffer));
			}
			usleep(50000);
		}
	}
	
	/**
	* @return void
	* @desc Verbindung wurde bgebrochen - aufraeumen
	*/
	private function disconnect() {
		$this->log('Verbindung wurde beendet.');
		$this->connection = null;
		$this->signals[] = array(IRC_QUIT, 'nick' => $this->actualNick);
	}
	
	/**
	* @return arary
	* @param string $string	IRC-Server-Message
	* @desc Parsen einer ICR-Server-Message
	*/
	private function parseMessage(		$string) {
		// RFC 1459 konforme RegExp um IRC-Nachrichten zu "parsen"
		
		// Fail: ":irc.inn.at.euirc.net 005 cow NETWORK=euIRCnet WATCH=128 SAFELIST PREFIX=(qaohv)*!@%+ CHANMODES=bewI,k,flBL,cimnprstACHKNOQRSVWXY CHANTYPES=#&+ KICKLEN=307 KNOCK MAP MAXBANS=60 MODES=6 NICKLEN=30 SILENCE=5 TOPICLEN=307 WALLCHOPS MAXCHANNELS=20 MAXTARGETS=20 :are available on this server (Zu viele parameter - laut RFC auch endlich viele erlaubt)
		// Fuers EuIRC non-RFC-Nick-Definitionen: 
		//  - Nick muss nicht mit einem Buchstaben anfangen
		//  - Erlaubt im Nick: |
		
		if (preg_match('/^(:(([\w]([\w\d-]?[\w\d])*)(\.([\w]([\w\d-]?[\w\d])*))*|([\w\d\[\]\\\\`^\{\}_\|-]*)(!([^\x20\x00\x0d\x0a@]+))?(@(([\w]([\w\d-]?[\w\d])*)(\.([\w]([\w\d-]?[\w\d])*))*))?)\s+)?(\w+|\d{3})(\s+([^:\x20\x00\x0d\x0a]*))?(\s+([^:\x20\x00\x0d\x0a]*))?(\s+([^:\x20\x00\x0d\x0a]*))?(\s+([^:\x20\x00\x0d\x0a]*))?(\s+([^:\x20\x00\x0d\x0a]*))?(\s+:([^\x00\x0d\x0a]*))?$/S', $string, $match)) {
			
			//var_dump($match);
			
			if (!empty($match[3])) {
				$msg['server'] = $match[3];
				$msg['nick'] = $match[3];
				$msg['ident'] = $match[3];
				$msg['host'] = $match[3];
			} else {
				$msg['nick'] = $match[8];
				$msg['ident'] = $match[10];
				$msg['host'] = $match[12];
				$msg['server'] = null;
			}
			
			$msg['type'] = $match[18];
			
			$msg['param'][] = @$match[20];
			$msg['param'][] = @$match[22];
			$msg['param'][] = @$match[24];
			$msg['param'][] = @$match[26];
			$msg['param'][] = @$match[28];
			
			$msg['text'] = @trim($match[30]);
			
			return $msg;
		} else {
			$this->log('Nicht zu parsende Nachricht: "' . trim($string) . '"');
			return array();
		}
	}
	
	/**
	* @return void
	* @desc Verbindung ueberpruefen, Nachrichten holen und parsen
	*/
	public function listen() {
		if ($this->timedOut()) $this->disconnect();
		if (@feof($this->connection)) $this->disconnect();
		if (!$this->connection) {
			sleep(KBOT_RECONNECT);
			$this->connect();
		}
		
		$this->readMessages();
		$this->handleMessages();
	}
	
	/**
	* @return void
	* @param string $source
	* @param string $type
	* @desc Antworten auf CTCP-Requests
	*/
	private function ctcp(				$source,
										$type) {
		switch (strtolower($type)) {
			case 'version':
				$this->notice($source, chr(1) . 'VERSION ' . KBOT_VERSION . ' Distributed under the GNU GPL; ' . @trim(`uname -ormp`) . chr(1));
			break;
			case 'finger':
				$this->notice($source, chr(1) . 'FINGER ' . str_replace('root', 'bot', trim(`whoami`)) . ': ' . trim(`uptime`) . chr(1));
			break;
			case 'source':
				$this->notice($source, chr(1) . 'SOURCE http://kore-nordmann.de/' . chr(1));
				$this->notice($source, chr(1) . 'SOURCE' . chr(1));
			break;
		}
	}
	
	/**
	* @return string
	* @desc Gibt den aktuellen Nick des Bots zurueck
	*/
	public function getActualNick() {
		return $this->actualNick;
	}
	
	/**
	* @return void
	* @desc Verarbeitet IRC-Nachrichten und formt die richtigen Signale daraus
	*/
	private function handleMessages() {
		foreach ($this->messages as $message) {
			switch (@strtoupper($message['type'])) {
				case '':
//					var_dump($message);
				break;
				case '001':
					// verbindung erfolgreich
					$this->signals[] = array(IRC_CONNECT);
				break;
				case '353': // Namen fuer beigetrenem Channel
					// Statustransformation
					$status = array(	'+' =>	IRC_VOICE,
										'%' =>	IRC_HALFOP,
										'@' =>	IRC_OP,
										'!' =>	IRC_ADMIN,
										'*' =>	IRC_FOUNDER);
				
					// Nicks voneinander trennen
					$nicks = explode(' ', $message['text']);
					
					// Nicks abarbeiten
					foreach ($nicks as $nick) {
						if (in_array($nick{0}, array_keys($status))) {
							// Nick hat Status - beides absetzen
							$this->signals[] = array(	IRC_JOIN, 
														'nick' =>		substr($nick, 1), 
														'channel' =>	$message['param'][2]);
							$this->signals[] = array(	IRC_GETMODE, 
														'nick' =>		substr($nick, 1), 
														'mode' =>		$status[$nick{0}], 
														'channel' =>	$message['param'][2]);
						} else {
							$this->signals[] = array(	IRC_JOIN, 
														'nick' =>		$nick, 
														'channel' =>	$message['param'][2]);
						}
					}
				break;
				case '431': // No nickname given
				case '432': // Erroneus nickname
				case '433': // Nickname is already in use
				case '436': // Nickname collision KILL
					$this->actualNick = $this->actualNick . '_';
					$this->nick($this->actualNick);
					$this->log('Nick schon in Gebrauch, neuer Nick: ' . $this->actualNick);
				break;
				case 'MODE':
					// Statustransformationen
					$status = array(	'v' =>	IRC_VOICE,
										'h' =>	IRC_HALFOP,
										'o' =>	IRC_OP,
										'a' =>	IRC_ADMIN,
										'q' =>	IRC_FOUNDER);

					// Splitten der Parameter und absetzen der Signale
					for ($i = 1; $i < strlen($message['param'][1]); $i++) {
						if (isset($status[$message['param'][1]{$i}])) {
							$this->signals[] = array(	($message['param'][1]{0} == '+' ? IRC_GETMODE : IRC_LOSEMODE), 
														'nick' =>		$message['param'][$i + 1],
														'mode' =>		$status[$message['param'][1]{$i}], 
														'channel' =>	$message['param'][0]);
						}
					}
				break;
				case 'KICK':
					// Signal fuer den kick absetzen
					$this->signals[] = array(	IRC_KICK, 
												'nick' =>		$message['nick'], 
												'ident' =>		$message['ident'], 
												'host' =>		$message['host'], 
												'kicked' =>		$message['param'][1], 
												'channel' =>	$message['param'][0]);
					// Der gekickte Nutzer verlaesst den Channel, also Part absetzen
					$this->signals[] = array(	IRC_PART, 
												'nick' =>		$message['param'][1], 
												'channel' =>	$message['param'][0],
												'text' =>		$message['text']);
				break;
				case 'PART':
					// Signal setzen
					$this->signals[] = array(	IRC_PART, 
												'nick' =>		$message['nick'], 
												'ident' =>		$message['ident'], 
												'host' =>		$message['host'], 
												'channel' =>	$message['param'][0],
												'text' =>		$message['text']);
				break;
				case 'QUIT':
					// Signal setzen
					$this->signals[] = array(	IRC_QUIT, 
												'nick' =>		$message['nick'],
												'ident' =>		$message['ident'], 
												'host' =>		$message['host'],
												'text' =>		$message['text']);
				break;
				case 'NICK':
					// Nickchange des eigenen Nicks speichern
					if ($message['nick'] == $this->actualNick) $this->actualNick = $message['text'];
					$this->signals[] = array(	IRC_NICK, 
												'nick' =>		$message['nick'], 
												'ident' =>		$message['ident'], 
												'host' =>		$message['host'], 
												'newnick' =>	$message['text']);
				break;
				case 'JOIN':
					if ($message['nick'] != $this->actualNick) {
						// Fuer eigenen Nick gab es einen 353er also muss der JOIN nicht extra verarbeitet werden
						$this->signals[] = array(	IRC_JOIN, 
													'nick' =>		$message['nick'], 
													'ident' =>		$message['ident'], 
													'host' =>		$message['host'], 
													'channel' =>	$message['text']);
					}
				break;
				case 'PRIVMSG':
				case 'NOTICE':
					// CTCP-Nachrichten erkennen und verarbeiten
					if (ord($message['text']{0}) == 1) {
						// CTCP
						$this->ctcp($message['nick'], substr($message['text'], 1, -1));						
						
					// Ist die nachricht ein Kommando?
					} elseif ($message['text']{0} == KBOT_COMMAND_CHAR) {
						// Kommando extrahieren
						if (($end = strpos($message['text'], ' ', 1)) === false) {
							$command = substr($message['text'], 1);
							$message['text'] = '';
						} else {
							$command = strtolower(substr($message['text'], 1, $end - 1));
							$message['text'] = trim(substr($message['text'], $end + 1));
						}
						// Signal setzen
						$this->signals[] = array(	IRC_COMMAND, 
													'nick' =>		$message['nick'], 
													'ident' =>		$message['ident'], 
													'host' =>		$message['host'], 
													'channel' =>	$message['param'][0], 
													'command' =>	$command, 
													'text' =>		$message['text']);
					
					// SOnst ist es "normaler" Text
					} else {
						// Signal setzen
						$this->signals[] = array(	IRC_MSG, 
													'nick' =>		$message['nick'],
													'ident' =>		$message['ident'], 
													'host' =>		$message['host'], 
													'channel' =>	$message['param'][0], 
													'text' =>		$message['text']);
					}
				break;
			}
		}
		// Nachrichtenarray wurde abgearbeitet, und kann geleert werden
		$this->messages = array();
	}
}

?>